"""
Generic processors that can be used in any module in the packages.

Examples:
    * Filtering down a collection items based on some predefined rules
    * Sorting a collection of items based on some predefined rules
    * Matching a collection of objects to some predefined rules
    * Limiting a collection of items based on some predefined rules
"""
