"""
Core functionality for all remote music streaming services.

Defines framework for interacting with music streaming service provided APIs
with some implementations for the handling of this data.

Also defines abstract classes to represent objects derived from the types of
responses that can be returned by the API.
"""
from ._response import RemoteResponse
