"""
All remote classes and operations with remote objects, files, tracks, albums, playlists, libraries etc.
"""
from musify.libraries.remote.spotify import SOURCE_NAME as SPOTIFY

REMOTE_SOURCES = frozenset([SPOTIFY])
